@@ -1,7 +1,7 @@
 from time import sleep
 from typing import Final, Type, TypeVar
 
-from serial.serialposix import Serial
+from serial import Serial
 from tabulate import tabulate
 
 CMD_RESET: Final = 0x44
@@ -31,10 +31,10 @@ ERROR_CODES = {
     0x04: 'Jam on the path between EJT Sensor and EXIT Sensor',
     0x05: 'A note Staying in EXIT Sensor',
     0x06: 'Ejecting the note suspected as rejected',
-    0x07: 'Abnormal note management (Flow Processing Error Inside)',
-    0x08: 'Abnormal note management (Flow Processing Error Inside)',
-    0x09: 'Abnormal note management (Flow Processing Error Inside)',
-    0x0A: 'Abnormal note management (Flow Processing Error Inside)',
+    0x07: 'Note count mis - match on eject sensor due to unexpected reason',
+    0x08: 'The note which should be rejected is passed on eject sensor',
+    0x09: 'The media length on eject sensor is too long due to slip or abnormal reason',
+    0x0A: 'The media length on exit sensor is too long due to slip or abnormal reason',
     0x0B: 'Detecting notes on the path before start of pick-up',
     0x0C: 'Dispensing too many notes for one transaction (Default limit: 100 notes including the rejected)',
     0x0D: 'Rejecting too many notes for one transaction (Default limit: 10 notes)',
@@ -48,6 +48,15 @@ ERROR_CODES = {
     0x26: 'Detecting no reject tray before start or for operation',
     0x30: 'Recognizing abnormal command',
     0x31: 'Recognizing abnormal parameter on the command',
+    0x32: 'Not to Operate VERIFY Command after Downloading and Reset',
+    0x33: 'Program area writing Failure',
+    0x34: 'Verify Failure',
+    0x35: 'EEPROM Write Failure',
+    0x36: 'Check Sum Error on Writing EEPROM',
+    0x40: 'During dispensing from the 2nd, 3rd or Bottom Cassette, the banknote coming from the Top Cassette is detected',
+    0x41: 'During dispensing from the 1st, 3rd or Bottom Cassette, the banknote coming from the 2nd Cassette is detected',
+    0x42: 'During dispensing from the 1st, 2nd or Bottom Cassette, the banknote coming from the 3rd Cassette is detected',
+    0x43: 'During dispensing from the 1st, 2nd or 3rd Cassette, the banknote coming from the 4th Cassette is detected',
     ACK_TIMEOUT: 'ACK timed out',
     RESPONSE_TIMEOUT: 'Response timed out',
     UNKNOWN_ERROR: 'Unknown Error',
@@ -386,10 +395,10 @@ class SensorDiagnosticsCommand(CommandMessage):
     _response_size = 6 + 5
 
     def __init__(self, *args):
-        super().__init__(*args)
         assert len(args) > 0
         assert args[0] in (0x31, 0x32, 0x33, 0x34)
         self.POS = args[0]
+        super().__init__(*args)
 
     def _compute_bcc(self, *_):
         super()._compute_bcc(self.POS)
@@ -488,6 +497,7 @@ class GetBillOpacitiesResponse(ResponseMessage):
 
     def __init__(self, *response: int):
         (
+            self.ERROR,
             self.OPAC1_HIGH, self.OPAC1_LOW, self.OPAC2_HIGH, self.OPAC2_LOW,
             self.OPAC3_HIGH, self.OPAC3_LOW, self.OPAC4_HIGH, self.OPAC4_LOW
         ) = response[4:-2]
@@ -495,6 +505,7 @@ class GetBillOpacitiesResponse(ResponseMessage):
 
     def _compute_bcc(self, *_):
         super()._compute_bcc(
+            self.ERROR,
             self.OPAC1_HIGH, self.OPAC1_LOW, self.OPAC2_HIGH, self.OPAC2_LOW,
             self.OPAC3_HIGH, self.OPAC3_LOW, self.OPAC4_HIGH, self.OPAC4_LOW
         )
@@ -546,7 +557,7 @@ class SetBillDispenseOrderResponse(ResponseMessage):
 class GetBillDispenseOrderCommand(CommandMessage):
     """The command will get the bill dispense order data."""
     CMD = CMD_GET_BILL_DISPENSE_ORDER
-    _response_size = 6 + 4
+    _response_size = 6 + 1 + 4
 
 
 class GetBillDispenseOrderResponse(ResponseMessage):
@@ -558,11 +569,11 @@ class GetBillDispenseOrderResponse(ResponseMessage):
     ORDER4: int  # The cassette location (type) that is last to be picked up. 0x31~0x34
 
     def __init__(self, *response: int):
-        (self.ORDER1, self.ORDER2, self.ORDER3, self.ORDER4) = response[4:-2]
+        (self.ERROR, self.ORDER1, self.ORDER2, self.ORDER3, self.ORDER4) = response[4:-2]
         super().__init__(*response)
 
     def _compute_bcc(self, *_):
-        super()._compute_bcc(self.ORDER1, self.ORDER2, self.ORDER3, self.ORDER4)
+        super()._compute_bcc(self.ERROR, self.ORDER1, self.ORDER2, self.ORDER3, self.ORDER4)
 
 
 class SetBillLengthsCommand(CommandMessage):
@@ -639,6 +650,7 @@ class GetBillLengthsResponse(ResponseMessage):
 
     def __init__(self, *response: int):
         (
+            self.ERROR,
             self.LENG1_HIGH, self.LENG1_LOW, self.LENG2_HIGH, self.LENG2_LOW,
             self.LENG3_HIGH, self.OPAC3_LOW, self.OPAC4_HIGH, self.OPAC4_LOW
         ) = response[4:-2]
@@ -646,6 +658,7 @@ class GetBillLengthsResponse(ResponseMessage):
 
     def _compute_bcc(self, *_):
         super()._compute_bcc(
+            self.ERROR,
             self.LENG1_HIGH, self.LENG1_LOW, self.LENG2_HIGH, self.LENG2_LOW,
             self.LENG3_HIGH, self.OPAC3_LOW, self.OPAC4_HIGH, self.OPAC4_LOW
         )
@@ -729,11 +742,17 @@ class PuloonLCDM4000:
         return self._run(PurgeCommand, PurgeResponse)
 
     def dispense(self, qty1: int = 0, qty2: int = 0, qty3: int = 0, qty4: int = 0,
-                 to1: int = 0x20, to2: int = 0x20) -> DispenseResponse:
+                 to1: bool = False, to2: int = 0) -> DispenseResponse:
         assert all(0 <= q <= 40 for q in (qty1, qty2, qty3, qty4)), "Quantity should be in range of 0 and 40"
         assert 0 < sum((qty1, qty2, qty3, qty4)) <= 40, "Quantity should be at least 1 and maximum 40 in total"
-        args = [q + 0x20 for q in (qty1, qty2, qty3, qty4)]
-        return self._run(DispenseCommand, DispenseResponse, *args, to1, to2)
+        assert 0 <= to2 <= 9, "Timeout should be in range of 0 and 9"
+        return self._run(
+            DispenseCommand,
+            DispenseResponse,
+            *(q + 0x20 for q in (qty1, qty2, qty3, qty4)),
+            0x1C if to1 else 0x20,
+            to2 + 0x30 if to1 else 0x20
+        )
 
     def test_dispense(self, qty1: int = 0, qty2: int = 0, qty3: int = 0, qty4: int = 0,
                       to1: bool = False, to2: int = 0) -> TestDispenseResponse:
@@ -755,43 +774,45 @@ class PuloonLCDM4000:
         assert 1 <= pos <= 4, "Dispenser has only 4 cassette"
         return self._run(SensorDiagnosticsCommand, SensorDiagnosticsResponse, pos + 0x30)
 
-    def set_bill_opacities(self, opac1_high: int = 0x30, opac1_low: int = 0x30,
-                           opac2_high: int = 0x30, opac2_low: int = 0x30,
-                           opac3_high: int = 0x30, opac3_low: int = 0x30,
-                           opac4_high: int = 0x30, opac4_low: int = 0x30) -> SetBillOpacitiesResponse:
-        assert all(0x30 <= v <= 0x3F for v in (
+    def set_bill_opacities(self, opac1_high: int = 15, opac1_low: int = 1,
+                           opac2_high: int = 15, opac2_low: int = 1,
+                           opac3_high: int = 15, opac3_low: int = 1,
+                           opac4_high: int = 15, opac4_low: int = 1) -> SetBillOpacitiesResponse:
+        assert all(1 <= v <= 15 for v in (
             opac1_high, opac1_low, opac2_high, opac2_low,
             opac3_high, opac3_low, opac4_high, opac4_low
-        )), "Opacity value should be in range of 0x31 and 0x34"
-        return self._run(SetBillOpacitiesCommand, SetBillOpacitiesResponse,
-                         opac1_high, opac1_low, opac2_high, opac2_low,
-                         opac3_high, opac3_low, opac4_high, opac4_low)
+        )), "Opacity value should be in range of 1 and 15"
+        return self._run(SetBillOpacitiesCommand, SetBillOpacitiesResponse, *(
+            v + 0x30 for v in (opac1_high, opac1_low, opac2_high, opac2_low,
+                               opac3_high, opac3_low, opac4_high, opac4_low)
+        ))
 
     def get_bill_opacities(self) -> GetBillOpacitiesResponse:
         return self._run(GetBillOpacitiesCommand, GetBillOpacitiesResponse)
 
-    def set_bill_dispense_order(self, order1: int = 0x31, order2: int = 0x31,
-                                order3: int = 0x31, order4: int = 0x31) -> SetBillDispenseOrderResponse:
-        assert all(0x31 <= v <= 0x34 for v in (order1, order2, order3, order4)), \
-            "Order value should be in range of 0x31 and 0x34"
+    def set_bill_dispense_order(self, order1: int = 1, order2: int = 2,
+                                order3: int = 3, order4: int = 4) -> SetBillDispenseOrderResponse:
+        assert all(1 <= v <= 4 for v in (order1, order2, order3, order4)), \
+            "Order value should be in range of 1 and 4"
         assert len({order1, order2, order3, order4}) == 4, "Order values should be unique"
         return self._run(SetBillDispenseOrderCommand, SetBillDispenseOrderResponse,
-                         order1, order2, order3, order4)
+                         *(v + 0x30 for v in (order1, order2, order3, order4)))
 
     def get_bill_dispense_order(self) -> GetBillDispenseOrderResponse:
         return self._run(GetBillDispenseOrderCommand, GetBillDispenseOrderResponse)
 
-    def set_bill_lengths(self, leng1_high: int = 0x30, leng1_low: int = 0x30,
-                         leng2_high: int = 0x30, leng2_low: int = 0x30,
-                         leng3_high: int = 0x30, leng3_low: int = 0x30,
-                         leng4_high: int = 0x30, leng4_low: int = 0x30) -> SetBillLengthsResponse:
-        assert all(0x30 <= v <= 0x3F for v in (
+    def set_bill_lengths(self, leng1_high: int = 15, leng1_low: int = 1,
+                         leng2_high: int = 15, leng2_low: int = 1,
+                         leng3_high: int = 15, leng3_low: int = 1,
+                         leng4_high: int = 15, leng4_low: int = 1) -> SetBillLengthsResponse:
+        assert all(1 <= v <= 15 for v in (
             leng1_high, leng1_low, leng2_high, leng2_low,
             leng3_high, leng3_low, leng4_high, leng4_low
-        )), "Length value should be in range of 0x31 and 0x34"
-        return self._run(SetBillLengthsCommand, SetBillLengthsResponse,
-                         leng1_high, leng1_low, leng2_high, leng2_low,
-                         leng3_high, leng3_low, leng4_high, leng4_low)
+        )), "Length value should be in range of 1 and 15"
+        return self._run(SetBillLengthsCommand, SetBillLengthsResponse, *(
+            v + 0x30 for v in (leng1_high, leng1_low, leng2_high, leng2_low,
+                               leng3_high, leng3_low, leng4_high, leng4_low)
+        ))
 
     def get_bill_lengths(self) -> GetBillLengthsResponse:
         return self._run(GetBillLengthsCommand, GetBillLengthsResponse)
